"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const axios_1 = __importDefault(require("axios"));
const backtrace_service_1 = require("backtrace-service");
const utilities_1 = require("../utilities");
const log = backtrace_service_1.Logger.getLogger();
exports.DEFECT_TABLE = 'defect';
exports.CRASH_TABLE = 'crash';
exports.CORRELATION_TABLE = 'correlation';
const WILDCARD = '*';
function escapeRegExp(str) {
    return str.replace(/[.*+?^${}()|[\]\\]/g, '\\$&'); // $& means the whole matched string
}
function getCorrelationTargetUrl(baseUrl, token, project) {
    return `${baseUrl}?token=${token}&project=${project}`;
}
function meetsCriterion(objectToCheck, keyname, value) {
    return !value || (value && objectToCheck[keyname] === value);
}
class Correlate {
    constructor(config, db, metrics) {
        this.correlate = async (o) => {
            var _a, _b, _c;
            // Use provided crash or fetch it from coroner
            let crashObject;
            if (o.crashObject) {
                crashObject = Correlate.formatObjects({ [o.fingerprint]: o.crashObject }, (_a = o.options) === null || _a === void 0 ? void 0 : _a.matchColumn)[0];
                delete o.crashObject;
            }
            else {
                const crashResults = await this.queryForCrash(o);
                if (crashResults.error) {
                    log.error(`Error: "${crashResults.error.message}" from crash query for ${o.fingerprint} in ` +
                        `universe/project ${o.universe}/${o.crashProject}.`);
                    return;
                }
                const crashQueryResults = Correlate.unpack(crashResults.response);
                crashObject = Correlate.formatObjects(crashQueryResults, (_b = o.options) === null || _b === void 0 ? void 0 : _b.matchColumn)[0];
            }
            if (!crashObject) {
                log.warn(`No result from crash query for ${o.fingerprint} in universe/project ${o.universe}/${o.crashProject}.`);
                return;
            }
            // Ignore correlating crash objects that have a "defect" classifier (e.g. handles cts-to-cts WF integration)
            // Doing the check here rather than filtering in the crash query since WF is likely sending the error object already
            if ((_c = crashObject.classifiers) === null || _c === void 0 ? void 0 : _c.includes("defect")) {
                log.info(`Classifiers for crash ${o.fingerprint} contain the word "defect" - aborting correlation processing on this fingerprint.`);
                return;
            }
            /*
                Get defects related to this crash
            */
            if (!crashObject.buildVersions) {
                log.debug(`Adding wildcard to crash object`);
                crashObject.buildVersions = [WILDCARD];
            }
            const { reducedFiles, reducedFunctions } = this.applyUserOptions(crashObject, o.options);
            // this does the work of pulling correlated defects
            const defectsByBuild = await this.getDefectsByBuild(crashObject, o.universe, reducedFunctions, reducedFiles);
            // Store each build version of correlation separately
            let count = 0;
            crashObject.buildVersions.forEach(async (buildVersion) => {
                let defects = defectsByBuild[buildVersion] || [];
                let crashIdForBuild;
                try {
                    crashIdForBuild = await this.insertCrash({
                        fingerprint: crashObject.fingerprint,
                        build: buildVersion,
                    }, o);
                }
                catch (e) {
                    return log.error(e);
                }
                // reduce defects based on classifier match
                if (o.options.classifierMatch) {
                    defects = Correlate.getDefectsWithClassifierMatch(defects, crashObject);
                }
                // insert in to correlate table
                defects.forEach((defectToSubmit) => {
                    count++;
                    const crashInfo = { fingerprint: crashObject.fingerprint, id: crashIdForBuild };
                    const defectInfo = { fingerprint: defectToSubmit.fingerprint, id: defectToSubmit.rowid };
                    const { file_name, function_name } = defectToSubmit;
                    this.upsertCrashDefectCorrelation(crashInfo, defectInfo, o, buildVersion, file_name, function_name);
                });
            });
            // log and record outcome of correlation
            if (count === 0) {
                if (this.metrics) {
                    this.metrics.addMetric(`zero-correlated-defects-found`, 1);
                }
                log.info(`No correlation results to record for ${o.fingerprint} in universe/crashProject ` +
                    `${o.universe}/${o.crashProject}.`);
                return;
            }
            if (this.metrics) {
                this.metrics.addMetric(`correlated-defects-found`, count);
            }
            log.info(`Recording ${count} correlation results for ${o.fingerprint} in universe/crashProject ` +
                `${o.universe}/${o.crashProject}.`);
        };
        this.db = db;
        this.config = config;
        this.metrics = metrics;
    }
    static convertBuildVersionsToRegex(buildVersions) {
        // e.g. ["1.2.3", "1.0.0"] becomes "(1\.2\.3)|(1\.0\.0)"
        const formattedVersions = buildVersions.map(buildVersion => {
            const escaped = escapeRegExp(buildVersion);
            return "(" + escaped + ")";
        });
        return formattedVersions.join('|');
    }
    static queryForError(fp, matchColumn) {
        const query = {
            group: ['fingerprint'],
            fold: {
                callstack: [['head']],
                'callstack.files': [['head']],
                classifiers: [['head']],
            },
            filter: [
                {
                    timestamp: [['at-least', 1]],
                    fingerprint: [['equal', fp]],
                }
            ]
        };
        if (!matchColumn)
            return query;
        query.fold[matchColumn] = [["histogram"]];
        return query;
    }
    ;
    static queryForTags(fp) {
        return {
            group: ['fingerprint'],
            fold: {
                "fingerprint;issues;tags": [['head']],
            },
            filter: [
                {
                    timestamp: [['at-least', 1]],
                    fingerprint: [['equal', fp]],
                }
            ]
        };
    }
    ;
    static queryForDefects(buildVersions) {
        const query = {
            filter: [
                {
                    classifiers: [['contains', 'defect']],
                    timestamp: [['greater-than', 0]],
                },
            ],
            group: ['fingerprint'],
            fold: {
                callstack: [['head']],
                'callstack.files': [['head']],
                classifiers: [['head']],
                build: [["histogram"]],
                "error.message": [['head']],
            },
        };
        if (buildVersions && buildVersions.length === 1 && buildVersions[0] === "*") {
            return query;
        }
        if (buildVersions) {
            query.filter[0].build = [['regular-expression', Correlate.convertBuildVersionsToRegex(buildVersions)]];
        }
        return query;
    }
    ;
    static unpackTagResult(queryResponse) {
        var _a, _b, _c, _d, _e, _f, _g;
        return ((_g = (_f = (_e = (_d = (_c = (_b = (_a = queryResponse) === null || _a === void 0 ? void 0 : _a.data) === null || _b === void 0 ? void 0 : _b.response) === null || _c === void 0 ? void 0 : _c.values) === null || _d === void 0 ? void 0 : _d[0]) === null || _e === void 0 ? void 0 : _e[1]) === null || _f === void 0 ? void 0 : _f[0]) === null || _g === void 0 ? void 0 : _g[0]) || "";
    }
    /* Take a query response and map column and values to form neater crash objects. */
    static unpack(queryResponse) {
        const result = {};
        if (!queryResponse.values)
            return result;
        queryResponse.values.forEach(value => {
            const fp = value[0];
            const colValues = value[1];
            if (fp && Array.isArray(colValues) && !result[fp]) {
                result[fp] = {};
                queryResponse.columns.forEach((column, index) => {
                    const label = column[0];
                    result[fp][label] = colValues[index];
                });
            }
        });
        return result;
    }
    async queryForCrash(obj) {
        const { crashProject, crashProjectToken, fingerprint, queryUrl, options: { matchColumn }, } = obj;
        let url = getCorrelationTargetUrl(queryUrl, crashProjectToken, crashProject);
        const results = await axios_1.default.post(url, Correlate.queryForError(fingerprint, matchColumn));
        return results.data;
    }
    /* update the tags of the crash issue to indicate it has correlated defects */
    async updateCrashTags(queryUrl, crashFingerprint, crashProjectToken, crashProject) {
        let existingTags;
        const url = `${queryUrl}?token=${crashProjectToken}&project=${crashProject}`;
        try {
            const existingTagsRawResult = await axios_1.default.post(url, Correlate.queryForTags(crashFingerprint));
            existingTags = Correlate.unpackTagResult(existingTagsRawResult);
            log.debug(`Fingerprint ${crashFingerprint} existingTags: ${existingTags}.`);
        }
        catch (e) {
            log.warn(`Error getting existing tags for fingerprint ${crashFingerprint}; skipping tag-setting step. ${e}`);
            return;
        }
        if (existingTags.includes(Correlate.DEFECTS_TAG)) {
            log.info(`Fingerprint ${crashFingerprint} already contains defect tag; skipping tag-setting step.`);
            return;
        }
        const newTags = `${existingTags} ${Correlate.DEFECTS_TAG}`.trim();
        const setTagsRequest = {
            "tags": {
                "action": "set",
                "value": newTags,
            },
            "fingerprint": [crashFingerprint],
            "project": crashProject,
        };
        const issuesUrl = `${queryUrl.replace(Correlate.apiQueryRegex, '/api/issue')}?token=${crashProjectToken}&project=${crashProject}`;
        try {
            await axios_1.default.post(issuesUrl, setTagsRequest);
            log.info(`Updated fingerprint ${crashFingerprint} tags to "${newTags}"`);
        }
        catch (e) {
            log.warn(`error updating correlation tag for fingerprint ${crashFingerprint}: ${e}`);
        }
    }
    async queryForDefects(obj, buildVersions) {
        const { defectProject, defectProjectToken, queryUrl } = obj;
        let url = getCorrelationTargetUrl(queryUrl, defectProjectToken, defectProject);
        const results = await axios_1.default.post(url, Correlate.queryForDefects(buildVersions));
        return results.data;
    }
    static formatObjects(unpackedObjects, matchColumn) {
        const fingerprints = [];
        for (let fp in unpackedObjects) {
            const obj = unpackedObjects[fp];
            let functions = [];
            if (obj['head(callstack)'] && obj['head(callstack)'].length > 0) {
                const parsed = JSON.parse(obj['head(callstack)'][0]);
                functions = parsed.frame && Array.isArray(parsed.frame) ? parsed.frame : [];
            }
            let files = [];
            if (obj['head(callstack.files)'] && obj['head(callstack.files)'].length > 0) {
                const parsed = JSON.parse(obj['head(callstack.files)'][0]);
                files = parsed.frame && Array.isArray(parsed.frame) ? parsed.frame : [];
            }
            const classifiers = obj['head(classifiers)'] ? obj['head(classifiers)'][0] : undefined;
            // Use matchColumn to set buildVersions
            let buildVersions;
            if (matchColumn && obj[`histogram(${matchColumn})`]) {
                log.debug(`object has matchColumn ${matchColumn}: ${obj[`histogram(${matchColumn})`]}`);
                const builds = obj[`histogram(${matchColumn})`];
                buildVersions = builds.map((build) => {
                    return build[0].toString(); // schema expects text
                });
            }
            const errorMessage = (obj['head(error.message)'] && obj['head(error.message)'].length > 0) ?
                obj['head(error.message)'][0] : undefined;
            // Set empty buildVersions as undefined to avoid "no column" error on defect query
            if (buildVersions && buildVersions.length === 1 && buildVersions[0] === undefined) {
                buildVersions = undefined;
            }
            fingerprints.push({
                fingerprint: fp,
                files,
                functions,
                classifiers,
                buildVersions,
                errorMessage
            });
        }
        return fingerprints;
    }
    static getDefectsWithClassifierMatch(defects, crash) {
        if (!crash.classifiers)
            return [];
        const crashClassifiers = crash.classifiers.split(' ');
        return defects.filter(defect => {
            return crashClassifiers.find(crashClassifier => {
                if (!defect.classifiers)
                    return false;
                return defect.classifiers.includes(crashClassifier);
            });
        });
    }
    async insertCrash(crashObject, o) {
        return this.db(exports.CRASH_TABLE)
            .first("rowid")
            .where({
            universe: o.universe,
            fingerprint: crashObject.fingerprint,
            project: o.crashProject,
            build_id: crashObject.build
        })
            .then((crashRowResult) => {
            if (crashRowResult && crashRowResult.rowid) {
                return crashRowResult.rowid;
            }
            return this.db(exports.CRASH_TABLE)
                .insert({
                universe: o.universe,
                fingerprint: crashObject.fingerprint,
                project: o.crashProject,
                build_id: crashObject.build
            })
                .then((rowIds) => {
                return rowIds[0];
            })
                .catch(e => log.error(e));
        })
            .catch(e => log.error(e));
    }
    // Used in "send" function to insert defects as they are ingested
    async insertDefect(defectObject, o) {
        return this.db(exports.DEFECT_TABLE)
            .first("rowid")
            .where({
            universe: o.universe,
            fingerprint: defectObject.fingerprint,
            project: o.defectProject,
            build_id: defectObject.build || WILDCARD,
        })
            .then((defectRowResult) => {
            if (defectRowResult && defectRowResult.rowid) {
                return defectRowResult.rowid;
            }
            return this.db(exports.DEFECT_TABLE)
                .insert({
                universe: o.universe,
                fingerprint: defectObject.fingerprint,
                project: o.defectProject,
                build_id: defectObject.build,
                error_message: defectObject.errorMessage,
                classifiers: defectObject.classifiers.join(" "),
                line_number: defectObject.lineNumber,
                function_name: defectObject.functionName,
                file_name: defectObject.fileName,
            })
                .then((rowIds) => {
                return rowIds[0];
            })
                .catch(e => log.error(e));
        })
            .catch(e => log.error(e));
    }
    async upsertCorrelation(crashRowId, defectRowId, startTime, options) {
        const fileName = options.fileName;
        const functionName = options.functionName;
        const whereValues = {
            crash: crashRowId,
            defect: defectRowId,
        };
        if (functionName) {
            whereValues.function_name = functionName;
        }
        if (fileName) {
            whereValues.file_name = fileName;
        }
        return this.db(exports.CORRELATION_TABLE)
            .first("rowid")
            .where(whereValues)
            .then((correlationResults) => {
            const latency = utilities_1.getLatency(startTime);
            if (correlationResults && correlationResults.rowid) {
                return this.db(exports.CORRELATION_TABLE)
                    .select({ rowid: correlationResults.rowid })
                    .update({ latency })
                    .then(_numberOfRowsUpdated => {
                    return { rowid: correlationResults.rowid, latency };
                });
            }
            return this.db(exports.CORRELATION_TABLE)
                .insert(Object.assign(Object.assign({}, whereValues), { latency }))
                .then((rowIds) => {
                return { rowid: rowIds[0], latency };
            });
        });
    }
    async upsertCrashDefectCorrelation(crashInfo, defectInfo, correlateRequest, build, fileName, functionName) {
        const { crashProject, crashProjectToken, queryUrl } = correlateRequest;
        const { fingerprint: crashFingerprint, id: crashId } = crashInfo;
        const { fingerprint: defectFingerprint, id: defectId } = defectInfo;
        try {
            const result = await this.upsertCorrelation(crashId, defectId, correlateRequest.startTime, {
                build,
                fileName,
                functionName,
            });
            if (this.metrics) {
                this.metrics.addMetric(`store-correlation`, 1);
            }
            log.info(`Stored correlation of crash/build ${crashFingerprint}/${build} and defect ${defectFingerprint}` +
                ` with latency of ${result.latency}`);
            this.updateCrashTags(queryUrl, crashFingerprint, crashProjectToken, crashProject);
        }
        catch (err) {
            if (this.metrics) {
                this.metrics.addMetric(`failed-store-correlation`, 1);
            }
            log.error(`Unable to store correlation of crash/build ${crashFingerprint}/${build} and defect ${defectFingerprint}`);
            return;
        }
    }
    async getDefectsByBuild(crashObject, universe, reducedFunctions = crashObject.functions, reducedFiles = crashObject.files) {
        const defectsByBuild = {};
        if (!crashObject.buildVersions || crashObject.buildVersions.length === 0) {
            crashObject.buildVersions = [WILDCARD];
        }
        try {
            await Promise.all(crashObject.buildVersions.map(async (build) => {
                if (build === WILDCARD) { // get all, disregarding build_id value
                    try {
                        const defectDataResults = await this.db(exports.DEFECT_TABLE)
                            .select('rowid', 'fingerprint', 'file_name', 'function_name', 'classifiers', "build_id").where({ universe }) // only select defects in the same universe
                            .whereIn('file_name', reducedFiles)
                            .andWhere((builder) => builder.whereIn('function_name', reducedFunctions)
                            .orWhereNull('function_name'));
                        defectsByBuild[build] = defectDataResults;
                    }
                    catch (e) {
                        log.warn(`error getting defects for build ${build} ${e}`);
                    }
                }
                else {
                    try {
                        const defectDataResults = await this.db(exports.DEFECT_TABLE)
                            .select('rowid', 'fingerprint', 'file_name', 'function_name', 'classifiers', "build_id").where({
                            build_id: build,
                            universe,
                        })
                            .then(results => {
                            return results.filter(result => {
                                return reducedFiles.includes(result.file_name) ||
                                    reducedFunctions.includes(result.function_name);
                            });
                        });
                        defectsByBuild[build] = defectDataResults;
                    }
                    catch (e) {
                        log.warn(`error getting defects for build ${build} ${e}`);
                    }
                }
            }));
        }
        catch (e) {
            log.warn(`Couldn't get defects by build ${e}`);
        }
        return defectsByBuild;
    }
    applyUserOptions(crash, opts) {
        let reducedFunctions = crash.functions.filter((fn) => {
            if (opts.ignoreFunctions && Array.isArray(opts.ignoreFunctions)) {
                return !opts.ignoreFunctions.includes(fn);
            }
            return true;
        });
        let reducedFiles = crash.files.filter((fn) => {
            if (opts.ignoreFiles && Array.isArray(opts.ignoreFiles)) {
                return !opts.ignoreFiles.includes(fn);
            }
            return true;
        });
        if (opts.limitFunctions !== undefined) {
            reducedFunctions = reducedFunctions.slice(0, opts.limitFunctions);
        }
        if (opts.limitFiles !== undefined) {
            reducedFiles = reducedFiles.slice(0, opts.limitFiles);
        }
        return { reducedFiles, reducedFunctions };
    }
    getCorrelatedDefects(universe, crashProject, crashFingerprint, options) {
        const { build, fileName, functionName } = options;
        return this.db({ crash: exports.CRASH_TABLE, defect: exports.DEFECT_TABLE, correlation: exports.CORRELATION_TABLE })
            .select({
            fingerprint: 'defect.fingerprint',
            errorMessage: 'defect.error_message',
            build: 'defect.build_id',
            fileName: 'correlation.file_name',
            functionName: 'correlation.function_name',
            project: 'defect.project',
            classifiers: 'defect.classifiers',
            lineNumber: 'defect.line_number',
        })
            .from(exports.CRASH_TABLE)
            .leftJoin(exports.CORRELATION_TABLE, 'crash.rowid', '=', 'correlation.crash')
            .leftJoin(exports.DEFECT_TABLE, 'correlation.defect', '=', 'defect.rowid')
            .whereRaw(`crash.universe = ? AND  crash.project = ? AND crash.fingerprint = ?`, [universe, crashProject, crashFingerprint])
            .then(results => {
            return results.filter(result => {
                return meetsCriterion(result, 'build', build) &&
                    meetsCriterion(result, 'fileName', fileName) &&
                    meetsCriterion(result, 'functionName', functionName);
            });
        })
            .catch((err) => {
            log.error(err);
            throw new Error(`Unable to fetch correlated defects for crash fingerprint ${crashFingerprint} and build ${build} in ` +
                `universe/crashProject ${universe}/${crashProject}.`);
        });
    }
}
exports.Correlate = Correlate;
Correlate.apiQueryRegex = /\/api\/query/;
Correlate.DEFECTS_TAG = "has-defects";
exports.default = Correlate;
//# sourceMappingURL=Correlate.js.map