"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const plist = require('plist');
const child_process_1 = __importDefault(require("child_process"));
const fs_1 = __importDefault(require("fs"));
const path = __importStar(require("path"));
const backtrace_service_1 = require("backtrace-service");
const Adaptor_1 = require("./Adaptor");
class ClangAnalyzer extends Adaptor_1.Adaptor {
    constructor(options) {
        super(options, 'clang-analyzer');
        // For parsing HTML files
        this.HTML_KEYS = ["BUGDESC", "BUGFILE", "BUGLINE", "BUGCOLUMN", "FUNCTIONNAME", "BUGCATEGORY", "BUGTYPE"];
        this.BUGDESC = "BUGDESC";
        this.BUGFILE = "BUGFILE";
        this.BUGLINE = "BUGLINE";
        this.BUGCOLUMN = "BUGCOLUMN";
        this.FUNCTIONNAME = "FUNCTIONNAME";
        this.BUGCATEGORY = "BUGCATEGORY";
        this.BUGTYPE = "BUGTYPE";
        this.htmlPropertiesRegex = /(?<=<!-- )(.+) (.*?)(?= -->)/gm;
    }
    getValuesFromHtml(src) {
        const values = {};
        let arr;
        while ((arr = this.htmlPropertiesRegex.exec(src)) !== null) {
            const match = arr[0];
            const htmlKey = match.split(" ")[0];
            const htmlValue = match.slice(match.indexOf(htmlKey) + htmlKey.length + 1);
            if (this.HTML_KEYS.includes(htmlKey)) {
                values[htmlKey] = htmlValue;
            }
        }
        return values;
    }
    // Yields n raw reports (raw reports come from plist and html files)
    *readReports(dir, n) {
        var _a, _b, _c;
        let clangVersion;
        const files = fs_1.default.readdirSync(dir, { withFileTypes: true });
        const htmlFiles = [];
        const seenHtmls = {};
        let reports = [];
        for (const f of files) {
            const f_ = path.join(dir, f.name);
            if (f.isDirectory()) {
                yield* this.readReports(f_, n); // need to yield to call generators recursively
                continue;
            }
            if (path.extname(f_) === '.plist') {
                const stats = fs_1.default.statSync(f_);
                this.log.info(`plist report size: ${stats.size}`); // gain insight to see if it's worth stream parsing these as well.
                const plRaw = fs_1.default.readFileSync(f_, { encoding: 'utf8' });
                const plistParsed = plist.parse(plRaw);
                const idRegexResult = /report-(.*?)\.plist$/.exec(f_);
                if ((_c = (_b = (_a = plistParsed.diagnostics) === null || _a === void 0 ? void 0 : _a[0]) === null || _b === void 0 ? void 0 : _b.HTMLDiagnostics_files) === null || _c === void 0 ? void 0 : _c[0]) { // not all plist files contain diagnostic info
                    for (const diagnostic of plistParsed.diagnostics) { // plist files contain multiple defects, generate a report for each
                        if (!diagnostic.HTMLDiagnostics_files || !diagnostic.HTMLDiagnostics_files[0]) {
                            continue;
                        }
                        seenHtmls[diagnostic.HTMLDiagnostics_files[0]] = true; // have only ever seen 1 file in here
                        const html = path.join(dir, diagnostic.HTMLDiagnostics_files[0]);
                        // Check that the html file exists (in some cases it does not, which breaks debug UI and mismatches with scan-build total count of reports.)
                        try {
                            fs_1.default.accessSync(html, fs_1.default.constants.R_OK);
                            if (!clangVersion) {
                                clangVersion = plistParsed.clang_version;
                            }
                            reports.push({ diagnostic, html, 'id': idRegexResult[1], clangVersion, files: plistParsed.files });
                            if (reports.length >= n) {
                                yield reports;
                                reports = [];
                            }
                        }
                        catch (_d) {
                            this.log.info(`Skipped processing plist file ${f_} because html file ${diagnostic.HTMLDiagnostics_files[0]} does not exist.`);
                        }
                    }
                }
            }
            else if (path.extname(f_) === '.html') {
                let basename = path.basename(f_);
                if (basename.startsWith("report-")) {
                    htmlFiles.push(basename);
                }
            }
        }
        // Check if there are extra html files we didn't process through plist files.
        // Seems to happen when the defect path spans multiple files.
        for (const htmlFile of htmlFiles) {
            if (seenHtmls[htmlFile]) {
                continue;
            }
            // this.log.debug(`${htmlFile} not processed via plist, parsing html.`);
            const htmlPath = path.join(dir, htmlFile);
            // We only need to look at beginning of html file for info.
            const htmlReadSize = 8600;
            const buffer = Buffer.alloc(htmlReadSize, 0, 'utf8');
            const fd = fs_1.default.openSync(htmlPath, 'r');
            fs_1.default.readSync(fd, buffer, 0, htmlReadSize, 0);
            const valuesFromHtml = this.getValuesFromHtml(buffer.toString('utf8'));
            const idRegexResult = /report-(.*?)\.html$/.exec(htmlFile);
            const locFile = valuesFromHtml[this.BUGFILE];
            // Skip bad or empty html files.
            if (!locFile || !valuesFromHtml[this.FUNCTIONNAME]) {
                this.log.warn(`Critical information is missing, omitting report ${htmlPath}`);
                continue;
            }
            reports.push({ diagnostic: {
                    HTMLDiagnostics_files: [htmlFile],
                    type: valuesFromHtml[this.BUGTYPE],
                    issue_context: valuesFromHtml[this.FUNCTIONNAME],
                    description: valuesFromHtml[this.BUGDESC],
                    check_name: "",
                    category: valuesFromHtml[this.BUGCATEGORY],
                    location: {
                        line: valuesFromHtml[this.BUGLINE],
                        col: valuesFromHtml[this.BUGCOLUMN],
                    }
                }, html: htmlPath, 'id': idRegexResult[1], clangVersion, files: [locFile] });
            fs_1.default.closeSync(fd);
            if (reports.length >= n) {
                yield reports;
                reports = [];
            }
        }
        // return any reports that were not returned under the n-limit conditions above.
        yield reports;
    }
    // Untar the archive into tmp directory
    initData() {
        if (!this.options.config.tar) {
            throw new Error('Missing tar in config');
        }
        const tmpPrefix = path.join(this.options.config.storage.path || './tmp/', 'tmp_clang_analyzer_');
        this.tmpDir = fs_1.default.mkdtempSync(tmpPrefix);
        try {
            child_process_1.default.execFileSync(this.options.config.tar, ['-C', this.tmpDir, '-xzf', '-'], { input: this.options.data });
        }
        catch (e) {
            throw new backtrace_service_1.ApiError("Error: Unable to extract report.", 422);
        }
    }
    // Yield n fully-processed reports at a time
    *reportsGenerator(n) {
        this.initData();
        if (!this.tmpDir) {
            this.log.warn("Something went wrong with initData - no tmpDir created. Aborting.");
            return;
        }
        const readReportsGenerator = this.readReports(this.tmpDir, n);
        let readReportsGeneratorState = readReportsGenerator.next();
        while (!readReportsGeneratorState.done) {
            const rawReports = readReportsGeneratorState.value;
            const formattedReports = this.formatRawReports(rawReports);
            yield formattedReports;
            readReportsGeneratorState = readReportsGenerator.next();
        }
    }
    // Converts raw reports into reports ready for submission to coroner
    formatRawReports(rawReports) {
        // Loop over JSON files and convert to Reports
        return rawReports.map((n) => {
            const { id, files, clangVersion, diagnostic, html } = n;
            const reportIdentifier = `${id}-${clangVersion}-${files[0]}`;
            const uuid = this.getUuid(reportIdentifier);
            const locFile = files[0];
            const reducedFile = path.basename(locFile);
            const fingerprint = this.getFingerprint(`${reducedFile}-${diagnostic.location.line}-${diagnostic.issue_context}-${diagnostic.type}`);
            const typeFormatted = diagnostic.type.toLowerCase().replace(/\s/g, "-");
            const classifiers = this.getClassifiers(typeFormatted);
            const attachmentPath = html;
            const threads = {
                main: {
                    name: '1',
                    fault: true,
                    stack: [
                        {
                            line: diagnostic.location.line,
                            column: diagnostic.location.col,
                            sourceCode: reducedFile,
                            funcName: diagnostic.issue_context,
                            library: reducedFile,
                        }
                    ]
                }
            };
            const sourceCode = {
                [reducedFile]: {
                    type: 'html-attachment',
                    attachment: diagnostic.HTMLDiagnostics_files[0],
                    startLine: diagnostic.location.line,
                    startPos: diagnostic.location.col,
                    path: locFile.split('/').slice(-1)[0],
                }
            };
            const report = {
                uuid,
                timestamp: parseInt((Date.now() / 1000).toString(), 10),
                agent: 'sca',
                lang: 'sca',
                langVersion: '0.0.0',
                // annotations: ___, // TODO
                classifiers,
                attributes: {
                    uuid,
                    'tool': `clang-analyzer`,
                    'tool.version': clangVersion,
                    'sca.error_type': diagnostic.type,
                    'sca.error_category': diagnostic.category,
                    'severity': '1',
                    'error.message': diagnostic.description,
                    'error.extended': diagnostic.check_name,
                    '_mod_fingerprint': fingerprint,
                    'callstack.files': JSON.stringify({
                        frame: [locFile]
                    }),
                },
                threads,
                sourceCode,
                mainThread: 'main',
            };
            return { report, attachmentPath };
        });
    }
    close() {
        // TODO: implement in js
        child_process_1.default.execFileSync('rm', ['-rf', this.tmpDir]);
    }
}
exports.ClangAnalyzer = ClangAnalyzer;
//# sourceMappingURL=ClangAnalyzer.js.map